package org.jibble.epsgraphics;

import java.awt.*;
import java.awt.image.*;
import java.awt.image.renderable.*;
import java.awt.geom.*;
import java.text.*;
import java.awt.font.*;
import java.util.*;


/**
 * EpsGraphics2D is suitable for creating high quality EPS graphics for
 * use in documents and papers, and can be used just like a standard
 * Graphics2D object.
 *  <p>
 * Many Java programs use Graphics2D to draw stuff on the screen, and while
 * it is easy to save the output as a png or jpeg file, it is a little
 * harder to export it as an EPS for including in a document or paper.
 *  <p>
 * This class makes the whole process extremely easy, because you can use
 * it as if it's a Graphics2D object.  The only difference is that all of
 * the implemented methods create EPS output, which means the diagrams you
 * draw can be resized without leading to any of the jagged edges you may
 * see when resizing pixel-based images, such as jpeg and png files.
 *  <p>
 *   Example usage:
 *  <p>
 * <pre>    Graphics2D g = new EpsGraphics2D();
 *    g.setColor(Color.black);
 *    
 *    // Line thickness 2.
 *    g.setStroke(new BasicStroke(2.0f));
 *    
 *    // Draw a line.
 *    g.drawLine(10, 10, 50, 10);
 *
 *    // Fill a rectangle in blue
 *    g.setColor(Color.blue);
 *    g.fillRect(10, 0, 20, 20);
 *
 *    // Get the EPS output.
 *    String output = g.toString();
 *    
 *    // Free up resources.
 *    g.dispose();</pre>
 *  <p>
 * You do not need to worry about the size of the canvas when drawing on a
 * EpsGraphics2D object.  The bounding box of the EPS document will
 * automatically resize to accomodate new items that you draw.
 *  <p>
 * Not all methods are implemented yet.  Those that are not are clearly
 * labelled.
 *  <p>
 * Copyright Paul Mutton,
 *           <a href="http://www.jibble.org/">http://www.jibble.org/</a>
 * 
 * @author  $Author: pjm2 $
 * @version $Id: EpsGraphics2D.java,v 1.10 2003/06/30 22:58:03 pjm2 Exp $
 */
public class EpsGraphics2D extends java.awt.Graphics2D {
    
    
    public static final String VERSION = "0.1.1";
    
    
    /**
     * Constructs a new EPS document that is initially empty and can be
     * drawn on like a Graphics2D object.
     */
    public EpsGraphics2D() {
        clear();
    }
    
    
    /**
     * This method is called to indicate that a particular method is not
     * supported yet.  The stack trace was printed to the standard output.
     */
    private void methodNotSupported() {
        EpsException e = new EpsException("Method not currently supported by EpsGraphics2D version " + VERSION);
        //e.printStackTrace(System.out);
    }
    
    
    /////////////// Specialist methods ///////////////////////
    
    
    /**
     * Clears the contents of the current EPS document.
     */
    public synchronized void clear() {
        _headers = new StringBuffer();
        _contents = new StringBuffer();
        setColor(getColor());
        setFont(getFont());
        setStroke(getStroke());
    }
    
    
    /**
     * Returns the title of the EPS document.
     */
    public String getTitle() {
        return _title;
    }
    
    
    /**
     * Sets the title of the EPS document.
     */
    public void setTitle(String title) {
        _title = title;
    }    
    
    
    /**
     * Returns the point after it has been transformed by the transformation.
     */
    private Point2D transform(float x, float y) {
        Point2D result = new Point2D.Float(x, y);
        result = _transform.transform(result, result);
        result.setLocation(result.getX(), -result.getY());
        return result;
    }
    
    
    /**
     * Updates the bounds of the current EPS document.
     */
    private void updateBounds(double x, double y) {
        if (x > maxX) {
            maxX = (float) x;
        }
        if (x < minX) {
            minX = (float) x;
        }
        if (y > maxY) {
            maxY = (float) y;
        }
        if (y < minY) {
            minY = (float) y;
        }
    }
    
    
    /**
     * Appends a line to the EPS document.  A new line character is added
     * to the end of the line when it is added.
     */
    private void append(String line) {
        _contents.append(line + "\n");
    }
    
    
    /**
     * Appends the commands required to draw a shape on the EPS document.
     */
    private void draw(Shape s, boolean filled) {
        // Update the bounds.
        Rectangle2D rect = s.getBounds2D();
        float lineRadius = _stroke.getLineWidth() / 2;
        float minX = (float) rect.getMinX() - lineRadius;
        float minY = (float) rect.getMinY() - lineRadius;
        float maxX = (float) rect.getMaxX() + lineRadius;
        float maxY = (float) rect.getMaxY() + lineRadius;
        Point2D min = transform(minX, minY);
        Point2D max = transform(maxX, maxY);
        updateBounds(min.getX(), min.getY());
        updateBounds(max.getX(), max.getY());
        
        //append("newpath");
        int type = 0;
        float[] coords = new float[6];
        PathIterator it = s.getPathIterator(null);
        while (!it.isDone()) {
            type = it.currentSegment(coords);
            Point2D point1 = transform(coords[0], coords[1]);
            Point2D point2 = transform(coords[2], coords[3]);
            Point2D point3 = transform(coords[4], coords[5]);
            float x1 = (float) point1.getX();
            float y1 = (float) point1.getY();
            float x2 = (float) point2.getX();
            float y2 = (float) point2.getY();
            float x3 = (float) point3.getX();
            float y3 = (float) point3.getY();
            
            if (type == PathIterator.SEG_CLOSE) {
                append("closepath");
            }
            else if (type == PathIterator.SEG_CUBICTO) {
                append(x1 + " " + y1 + " " + x2 + " " + y2 + " " + x3 + " " + y3 + " curveto");
            }
            else if (type == PathIterator.SEG_LINETO) {
                append(x1 + " " + y1 + " lineto");
            }
            else if (type == PathIterator.SEG_MOVETO) {
                append(x1 + " " + y1 + " moveto");
            }
            else if (type == PathIterator.SEG_QUADTO) {
                append(x1 + " " + y1 + " " + x1 + " " + y1 + " " + x2 + " " + y2 + " curveto");
            }
            else if (type == PathIterator.WIND_EVEN_ODD) {
                // Ignore.
            }
            else if (type == PathIterator.WIND_NON_ZERO) {
                // Ignore.
            }
            it.next();
        }
        //append("closepath");
        if (filled) {
            append("fill");
        }
        else {
            append("stroke");
        }
    }
    
    
    /**
     * Returns a hex string that always contains two characters.
     */
    private String toHexString(int n) {
        String result = Integer.toString(n, 16);
        while (result.length() < 2) {
            result = "0" + result;
        }
        return result;
    }
    
    
    /////////////// Graphics2D methods ///////////////////////
    
    
    /**
     * Draws a 3D rectangle outline.  If it is raised, light appears to come
     * from the top left.
     */
    public void draw3DRect(int x, int y, int width, int height, boolean raised) {
        Color originalColor = getColor();
        Stroke originalStroke = getStroke();
        
        setStroke(new BasicStroke(1.0f));
        
        if (raised) {
            setColor(originalColor.brighter());
        }
        else {
            setColor(originalColor.darker());
        }
        
        drawLine(x, y, x + width, y);
        drawLine(x, y, x, y + height);
        
        if (raised) {
            setColor(originalColor.darker());
        }
        else {
            setColor(originalColor.brighter());
        }
        
        drawLine(x + width, y + height, x, y + height);
        drawLine(x + width, y + height, x + width, y);
        
        setColor(originalColor);
        setStroke(originalStroke);
    }


    /**
     * Fills a 3D rectangle.  If raised, it has bright fill and light appears
     * to come from the top left.
     */
    public void fill3DRect(int x, int y, int width, int height, boolean raised) {
        Color originalColor = getColor();
        
        if (raised) {
            setColor(originalColor.brighter());
        }
        else {
            setColor(originalColor.darker());
        }
        draw(new Rectangle(x, y, width, height), true);
        setColor(originalColor);
        draw3DRect(x, y, width, height, raised);
    }
    
    
    /**
     * Draws a Shape on the EPS document.
     */
    public void draw(Shape s) {
        draw(s, false);
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public boolean drawImage(Image img, AffineTransform xform, ImageObserver obs) {
        methodNotSupported();
        return false;
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void drawImage(BufferedImage img, BufferedImageOp op, int x, int y) {
        methodNotSupported();
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void drawRenderedImage(RenderedImage img, AffineTransform xform) {
        methodNotSupported();
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void drawRenderableImage(RenderableImage img, AffineTransform xform) {
        methodNotSupported();
    }
    
    
    /**
     * Draws a string at (x,y)
     */
    public void drawString(String str, int x, int y) {
        drawString(str, (float) x, (float) y);
    }
    
    
    /**
     * Draws a string at (x,y)
     */
    public void drawString(String s, float x, float y) {
        char[] chars = s.toCharArray();
        StringBuffer buffer = new StringBuffer();
        for (int i = 0; i < chars.length; i++) {
            char ch = chars[i];
            if (ch == '\n') {
                buffer.append("\\n");
            }
            else if (ch == '\r') {
                buffer.append("\\r");
            }
            else if (ch == '\t') {
                buffer.append("\\t");
            }
            else if (ch == '\b') {
                buffer.append("\\b");
            }
            else if (ch == '\f') {
                buffer.append("\\f");
            }
            else if (ch == '\\') {
                buffer.append("\\\\");
            }
            else if (ch == '(') {
                buffer.append("\\(");
            }
            else if (ch == ')') {
                buffer.append("\\)");
            }
            else {
                buffer.append(ch);
            }
        }
        Point2D point = transform(x, y);
        x = (float) point.getX();
        y = (float) point.getY();
        append(x + " " + y + " moveto");
        append("(" + buffer.toString() + ") show");
        updateBounds(x, y);
    }
    
    
    /**
     * Draws the characters of an AttributedCharacterIterator, starting from
     * (x,y)
     */
    public void drawString(AttributedCharacterIterator iterator, int x, int y) {
        drawString(iterator, (float) x, (float) y);
    }
    
    
    /**
     * Draws the characters of an AttributedCharacterIterator, starting from
     * (x,y)
     */
    public void drawString(AttributedCharacterIterator iterator, float x, float y) {
        StringBuffer buffer = new StringBuffer();
        for (char ch = iterator.first(); ch != CharacterIterator.DONE; ch = iterator.next()) {
            buffer.append(ch);
        }
        drawString(buffer.toString(), x, y);
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void drawGlyphVector(GlyphVector g, float x, float y) {
        methodNotSupported();
    }
    
    
    /**
     * Fills a Shape on the EPS document.
     */
    public void fill(Shape s) {
        draw(s, true);
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public boolean hit(Rectangle rect, Shape s, boolean onStroke) {
        methodNotSupported();
        return false;
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public GraphicsConfiguration getDeviceConfiguration() {
        methodNotSupported();
        return null;
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void setComposite(Composite comp) {
        methodNotSupported();
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void setPaint(Paint paint) {
        methodNotSupported();
    }
    
    
    /**
     * Sets the stroke.  Only accepts BasicStroke objects (or subclasses of
     * BasicStroke).
     */
    public void setStroke(Stroke s) {
        if (s instanceof BasicStroke) {
            _stroke = (BasicStroke) s;
            
            append((_stroke.getLineWidth() * _transform.getScaleX()) + " setlinewidth");
            append((_stroke.getMiterLimit() * _transform.getScaleX()) + " setmiterlimit");
            append(_stroke.getLineJoin() + " setlinejoin");
            append(_stroke.getEndCap() + " setlinecap");
            
            StringBuffer dashes = new StringBuffer();
            dashes.append("[ ");
            float[] dashArray = _stroke.getDashArray();
            if (dashArray != null) {
                for (int i = 0; i < dashArray.length; i++) {
                    dashes.append((dashArray[i] * _transform.getScaleX()) + " ");
                }
            }
            dashes.append("]");
            append(dashes.toString() + " 0 setdash");
        }
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void setRenderingHint(RenderingHints.Key hintKey, Object hintValue) {
        methodNotSupported();
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public Object getRenderingHint(RenderingHints.Key hintKey) {
        methodNotSupported();
        return null;
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void setRenderingHints(Map hints) {
        methodNotSupported();
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void addRenderingHints(Map hints) {
        methodNotSupported();
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public RenderingHints getRenderingHints() {
        methodNotSupported();
        return null;
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void translate(int x, int y) {
        methodNotSupported();
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void translate(double tx, double ty) {
        methodNotSupported();
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void rotate(double theta) {
        methodNotSupported();
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void rotate(double theta, double x, double y) {
        methodNotSupported();
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void scale(double sx, double sy) {
        methodNotSupported();
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void shear(double shx, double shy) {
        methodNotSupported();
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void transform(AffineTransform Tx) {
        methodNotSupported();
    }
    
    
    /**
     * Sets the AffineTransform to be used by this EpsGraphics2D.
     */
    public void setTransform(AffineTransform Tx) {
        _transform = Tx;
        // Need to update the stroke and font so they know the scale changed
        setStroke(getStroke());
        setFont(getFont());
    }
    
    
    /**
     * Gets the AffineTransform used by this EpsGraphics2D.
     */
    public AffineTransform getTransform() {
        return _transform;
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public Paint getPaint() {
        methodNotSupported();
        return null;
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public Composite getComposite() {
        methodNotSupported();
        return null;
    }
    
    
    /**
     * Sets the background color to be used by the clearRect method.
     */
    public void setBackground(Color color) {
        _backgroundColor = color;
    }
    
    
    /**
     * Gets the background color that is used by the clearRect method.
     */
    public Color getBackground() {
        return _backgroundColor;
    }
    
    
    /**
     * Returns the Stroke currently used.  Guaranteed to be an instance of
     * BasicStroke.
     */
    public Stroke getStroke() {
        return _stroke;
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void clip(Shape s) {
        methodNotSupported();
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public FontRenderContext getFontRenderContext() {
        methodNotSupported();
        return null;
    }
    
    
    /////////////// Graphics methods ///////////////////////
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public Graphics create() {
        methodNotSupported();
        return null;
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public Graphics create(int x, int y, int width, int height) {
        methodNotSupported();
        return null;
    }
    
    
    /**
     * Returns the current Color.  This will be a default value (black)
     * until it is changed using the setColor method.
     */
    public Color getColor() {
        return _color;
    }
    
    
    /**
     * Sets the Color to be used when drawing all future shapes, text, etc.
     */
    public void setColor(Color c) {
        _color = c;
        append((c.getRed() / 255f) + " " + (c.getGreen() / 255f) + " " + (c.getBlue() / 255f) + " setrgbcolor");
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void setPaintMode() {
        methodNotSupported();
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void setXORMode(Color c1) {
        methodNotSupported();
    }
    
    
    /**
     * Returns the Font currently being used.
     */
    public Font getFont() {
        return _font;
    }
    
    
    /**
     * Sets the Font to be used in future text.
     */
    public void setFont(Font font) {
        _font = font;
        append("/" + _font.getFamily() + " findfont " + ((int) (_font.getSize() * _transform.getScaleX())) + " scalefont setfont");
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public FontMetrics getFontMetrics() {
        methodNotSupported();
        return null;
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public FontMetrics getFontMetrics(Font f) {
        methodNotSupported();
        return null;
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public Rectangle getClipBounds() {
        methodNotSupported();
        return null;
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void clipRect(int x, int y, int width, int height) {
        methodNotSupported();
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void setClip(int x, int y, int width, int height) {
        methodNotSupported();
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public Shape getClip() {
        methodNotSupported();
        return null;
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void setClip(Shape clip) {
        methodNotSupported();
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void copyArea(int x, int y, int width, int height, int dx, int dy) {
        methodNotSupported();
    }
    
    
    /**
     * Draws a straight line from (x1,y1) to (x2,y2).
     */
    public void drawLine(int x1, int y1, int x2, int y2) {
        Shape shape = new Line2D.Float(x1, y1, x2, y2);
        draw(shape);
    }
    
    
    /**
     * Fills a rectangle with top-left corner placed at (x,y).
     */
    public void fillRect(int x, int y, int width, int height) {
        Shape shape = new Rectangle(x, y, width, height);
        draw(shape, true);
    }
    
    
    /**
     * Draws a rectangle with top-left corner placed at (x,y).
     */
    public void drawRect(int x, int y, int width, int height) {
        Shape shape = new Rectangle(x, y, width, height);
        draw(shape);
    }
    
    
    /**
     * Clears a rectangle with top-left corner placed at (x,y) using the
     * current background color.
     */
    public void clearRect(int x, int y, int width, int height) {
        Color originalColor = getColor();
        
        setColor(getBackground());
        Shape shape = new Rectangle(x, y, width, height);
        draw(shape, true);
        
        setColor(originalColor);
    }
    
    
    /**
     * Draws a rounded rectangle.
     */
    public void drawRoundRect(int x, int y, int width, int height, int arcWidth, int arcHeight) {
        Shape shape = new RoundRectangle2D.Float(x, y, width, height, arcWidth, arcHeight);
        draw(shape);
    }
    
    
    /**
     * Fills a rounded rectangle.
     */
    public void fillRoundRect(int x, int y, int width, int height, int arcWidth, int arcHeight) {
        Shape shape = new RoundRectangle2D.Float(x, y, width, height, arcWidth, arcHeight);
        draw(shape, true);
    }
    
    
    /**
     * Draws an oval.
     */
    public void drawOval(int x, int y, int width, int height) {
        Shape shape = new Ellipse2D.Float(x, y, width, height);
        draw(shape);
    }
    
    
    /**
     * Fills an oval.
     */
    public void fillOval(int x, int y, int width, int height) {
        Shape shape = new Ellipse2D.Float(x, y, width, height);
        draw(shape, true);
    }
    
    
    /**
     * Draws an arc.
     */
    public void drawArc(int x, int y, int width, int height, int startAngle, int arcAngle) {
        Shape shape = new Arc2D.Float(x, y, width, height, startAngle, arcAngle, Arc2D.OPEN);
        draw(shape);
    }
    
    
    /**
     * Fills an arc.
     */
    public void fillArc(int x, int y, int width, int height, int startAngle, int arcAngle) {
        Shape shape = new Arc2D.Float(x, y, width, height, startAngle, arcAngle, Arc2D.PIE);
        draw(shape, true);
    }
    
    
    /**
     * Draws a polyline.
     */
    public void drawPolyline(int[] xPoints, int[] yPoints, int nPoints) {
        if (nPoints > 0) {
            GeneralPath path = new GeneralPath();
            path.moveTo(xPoints[0], yPoints[0]);
            for (int i = 1; i < nPoints; i++) {
                path.lineTo(xPoints[i], yPoints[i]);
            }
            draw(path);
        }
    }
    
    
    /**
     * Draws a polygon made with the specified points.
     */
    public void drawPolygon(int[] xPoints, int[] yPoints, int nPoints) {
        Shape shape = new Polygon(xPoints, yPoints, nPoints);
        draw(shape);
    }
    
    
    /**
     * Draws a polygon.
     */
    public void drawPolygon(Polygon p) {
        draw(p);
    }
    
    
    /**
     * Fills a polygon made with the specified points.
     */
    public void fillPolygon(int[] xPoints, int[] yPoints, int nPoints) {
        Shape shape = new Polygon(xPoints, yPoints, nPoints);
        draw(shape, true);        
    }
    
    
    /**
     * Fills a polygon.
     */
    public void fillPolygon(Polygon p) {
        draw(p, true);
    }
    
    
    /**
     * Draws the specified characters, starting from (x,y)
     */
    public void drawChars(char[] data, int offset, int length, int x, int y) {
        String string = new String(data, offset, length);
        drawString(string, x, y);
    }
    
    
    /**
     * Draws the specified bytes, starting from (x,y)
     */
    public void drawBytes(byte[] data, int offset, int length, int x, int y) {
        String string = new String(data, offset, length);
        drawString(string, x, y);
    }
    
    
    /**
     * Draws an image.
     */
    public boolean drawImage(Image img, int x, int y, ImageObserver observer) {
        return drawImage(img, x, y, Color.white, observer);
    }
    
    
    /**
     * Draws an image.
     */
    public boolean drawImage(Image img, int x, int y, int width, int height, ImageObserver observer) {
        return drawImage(img, x, y, width, height, Color.white, observer);
    }
    
    
    /**
     * Draws an image.
     */
    public boolean drawImage(Image img, int x, int y, Color bgcolor, ImageObserver observer) {
        int width = img.getWidth(null);
        int height = img.getHeight(null);
        return drawImage(img, x, y, width, height, bgcolor, observer);
    }
    
    
    /**
     * Draws an image.
     */
    public boolean drawImage(Image img, int x, int y, int width, int height, Color bgcolor, ImageObserver observer) {
        return drawImage(img, x, y, x + width, y + height, 0, 0, width, height, bgcolor, observer);
    }
    
    
    /**
     * Draws an image.
     */
    public boolean drawImage(Image img, int dx1, int dy1, int dx2, int dy2, int sx1, int sy1, int sx2, int sy2, ImageObserver observer) {
        return drawImage(img, dx1, dy1, dx2, dy2, sx1, sy1, sx2, sy2, Color.white, observer);
    }
    
    
    /**
     * Draws an image.
     */
    public boolean drawImage(Image img, int dx1, int dy1, int dx2, int dy2, int sx1, int sy1, int sx2, int sy2, Color bgcolor, ImageObserver observer) {
        if (dx1 >= dx2) {
            throw new IllegalArgumentException("dx1 >= dx2");
        }
        if (sx1 >= sx2) {
            throw new IllegalArgumentException("sx1 >= sx2");
        }
        if (dy1 >= dy2) {
            throw new IllegalArgumentException("dy1 >= dy2");
        }
        if (sy1 >= sy2) {
            throw new IllegalArgumentException("sy1 >= sy2");
        }

        append("gsave");

        int destWidth = dx2 - dx1;
        int destHeight = dy2 - dy1;
        System.err.println(destWidth + " " + destHeight);
        
        BufferedImage bImage = new BufferedImage(destWidth, destHeight, BufferedImage.TYPE_INT_RGB);
        Graphics graphics = bImage.getGraphics();
        graphics.drawImage(img, 0, 0, destWidth, destHeight, sx1, sy1, sx2, sy2, bgcolor, null);

        Point2D translateTo = transform(dx1, dy1);
        Point2D endAt = transform(dx2, dy2);
        Point2D scaleTo = transform(destWidth, destHeight);
        updateBounds(translateTo.getX(), translateTo.getY());
        updateBounds(endAt.getX(), endAt.getY());
        append(translateTo.getX() + " " + (translateTo.getY() + scaleTo.getY()) + " translate");
        append(scaleTo.getX() + " " + scaleTo.getY() + " scale");
        append(destWidth + " " + destHeight + " 8 [" + destWidth + " 0 0 " + destHeight + " 0 " + destHeight + "]");
        append("{<");
        StringBuffer line = new StringBuffer();
        for (int y = 0; y < destHeight; y++) {
            for (int x = 0; x < destWidth; x++) {
                Color color = new Color(bImage.getRGB(x, y));
                line.append(toHexString(color.getRed()) + toHexString(color.getGreen()) + toHexString(color.getBlue()));
                if (line.length() > 64) {
                    append(line.toString());
                    line = new StringBuffer();
                }
            }
        }
        if (line.length() > 0) {
            append(line.toString());
        }
        append(">}");
        append("false 3 colorimage");
        
        append("grestore");
        
        return true;
    }
    
    
    /**
     * Disposes of all resources used by this EpsGraphics2D object.
     */
    public void dispose() {
        clear();
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public void finalize() {
        super.finalize();
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public String toString() {
        _headers = new StringBuffer();
        
        _headers.append("%!PS-Adobe-3.0 EPSF-3.0\n");
        _headers.append("%%Creator: EpsGraphics2D " + VERSION + " by Paul Mutton, http://www.jibble.org/\n");
        _headers.append("%%Title: " + _title + "\n");
        _headers.append("%%CreationDate: " + new Date() + "\n");
        _headers.append("%%DocumentData: Clean7Bit\n");
        _headers.append("%%BoundingBox: " + ((int) minX - 1) + " " + ((int) minY - 1) + " " + ((int) maxX + 1) + " " + ((int) maxY + 1) + "\n");
        _headers.append("%%DocumentProcessColors: Black\n");
        _headers.append("%%ColorUsage: Color\n");
        _headers.append("%%Origin: " + ((int) minX - 1) + " " + ((int) minY - 1) + "\n");
        _headers.append("%%Pages: 1\n");
        _headers.append("%%Page: 1 1\n");
        _headers.append("%%EndComments\n\n");
        
        StringBuffer output = new StringBuffer();
        output.append(_headers);
        output.append(_contents);
        output.append("\n");
        output.append("%%EOF");
        return output.toString();
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public boolean hitClip(int x, int y, int width, int height) {
        methodNotSupported();
        return false;
    }
    
    
    /**
     * <b><i>Not implemented</i></b>
     */
    public Rectangle getClipBounds(Rectangle r) {
        methodNotSupported();
        return null;
    }
    
    
    private float minX = Float.POSITIVE_INFINITY;
    private float minY = Float.POSITIVE_INFINITY;
    private float maxX = Float.NEGATIVE_INFINITY;
    private float maxY = Float.NEGATIVE_INFINITY;
    
    private Color _color = Color.black;
    private Color _backgroundColor = Color.black;
    private BasicStroke _stroke = new BasicStroke();
    private Font _font = Font.decode(null);
    
    private String _title = "Untitled";
    private StringBuffer _headers = new StringBuffer();
    private StringBuffer _contents = new StringBuffer();
    
    private AffineTransform _transform = new AffineTransform();
    
}